# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
    Bulbs is a Python persistence framework for graph databases.
    
    Bulbs is a Python client library for Rexster, a REST server that 
    provides access to any Blueprints-enabled graph database, including 
    TinkerGraph, Neo4j, OrientDB, DEX, and Sail RDF Stores. 

    Home Page: http://bulbflow.com
 
    Copyright (c) 2011, James Thornton (http://jamesthornton.com)
    License: BSD (see LICENSE for details)
"""
__version__ = '0.2.1'
__author__ = 'James Thornton'
__license__ = 'BSD'
